/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.editingsupport;

import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;
import static org.fortiss.tooling.kernel.utils.KernelModelElementUtils.computeFullyQualifiedName;
import static org.fortiss.tooling.kernel.utils.KernelModelElementUtils.computeRelativeName;
import static org.fortiss.tooling.kernel.utils.KernelModelElementUtils.getParentElement;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EEnum;
import org.eclipse.emf.ecore.EEnumLiteral;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.fortiss.tooling.base.annotation.valueprovider.EStructuralFeatureDescriptor;
import org.fortiss.tooling.base.annotation.valueprovider.EStructuralFeatureDescriptor.EReferenceScope;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.kernel.model.IIdLabeled;
import org.fortiss.tooling.kernel.model.INamedElement;
import org.fortiss.tooling.kernel.model.IProjectRootElement;

/**
 * Translates between the labels shown in a GUI and actual model elements (=values) (e.g.,
 * required to integrate {@link EReference}s into {@link ComboBoxEditingSupport} or
 * {@link MultiSelectionEditingSupport}. This class ensures that its label set is unique.
 * 
 * @author barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 4BE6EEA47866CFDB0C4ED2ADEB442E69
 */
public class LabelValueMapping {

	/** Map: label -> actual value */
	private Map<String, Object> labelToValueMap;

	/** {@link String}-based input choice */
	Collection<String> stringLabels;

	/**
	 * Constructs a {@link LabelValueMapping}. {@code eStructuralFeatureDescriptor} may be
	 * {@code null} in which case this translator acts as the identity-function.
	 */
	public LabelValueMapping(EStructuralFeatureDescriptor eStructuralFeatureDescriptor,
			IAnnotatedSpecification specification, EObject modelElement,
			Collection<String> stringLabels) {

		if(eStructuralFeatureDescriptor == null) {
			this.stringLabels = stringLabels;
			return;
		}

		EClassifier eType = eStructuralFeatureDescriptor.getEType(specification);
		if(eType instanceof EEnum) {
			createEnumLabels((EEnum)eType, eStructuralFeatureDescriptor, modelElement);
		} else if(eType instanceof EClass) {

			EObject root = null;
			switch(eStructuralFeatureDescriptor.getEReferenceScope()) {
				case PROJECT:
					root = getParentElement(modelElement, IProjectRootElement.class, true);
					if(root != null) {
						// Get FileProject (type is not available in this plugin)
						root = root.eContainer();
					}
					break;
				case MODEL:
					root = getParentElement(modelElement, IProjectRootElement.class, true);
					break;
				case SUB_MODEL:
				case STRICT_SUB_MODEL:
					root = modelElement;
					break;
			}

			if(root == null) {
				return;
			}

			EList<EObject> values = new BasicEList<EObject>();
			if(eStructuralFeatureDescriptor.getEReferenceScope() == EReferenceScope.SUB_MODEL) {
				values.add(root);
			}
			@SuppressWarnings("unchecked") Class<? extends EObject> instanceClass =
					(Class<? extends EObject>)eType.getInstanceClass();
			for(EObject obj : getChildrenWithType(root, instanceClass)) {
				if(eStructuralFeatureDescriptor.isAvailableObject(obj, specification, modelElement)) {
					values.add(obj);
				}
			}

			labelToValueMap = new LinkedHashMap<String, Object>();
			for(Object v : values) {
				if(v instanceof INamedElement) {
					if(root instanceof INamedElement) {
						String label = computeRelativeName((INamedElement)root, (INamedElement)v);
						if(eStructuralFeatureDescriptor.getEReferenceScope().equals(
								EStructuralFeatureDescriptor.EReferenceScope.PROJECT)) {
							// Strip explicit qualification of project name
							int i = label.indexOf(".");
							if(i != -1) {
								label = label.substring(i + 1);
							}
						}
						addLabelValuePair(label, v);
					} else {
						addLabelValuePair(computeFullyQualifiedName((INamedElement)v), v);
					}
				} else {
					addLabelValuePair(v.toString(), v);
				}
			}
		}
	}

	/**
	 * Creates the enum labels based on the literals that are given by the
	 * {@link EStructuralFeatureDescriptor}.
	 */
	private void createEnumLabels(EEnum eType,
			EStructuralFeatureDescriptor eStructuralFeatureDescriptor, EObject modelElement) {
		labelToValueMap = new LinkedHashMap<String, Object>();
		Set<String> enumValues = new TreeSet<String>();
		Collection<EEnumLiteral> literalsToShow =
				eStructuralFeatureDescriptor.getEnumLiterals(eType, modelElement);
		for(EEnumLiteral e : literalsToShow) {
			enumValues.add(e.getName());
			addLabelValuePair(e.getName(), e.getInstance());
		}
	}

	/** Returns the set of labels managed by this {@link LabelValueMapping}. */
	public Collection<String> getLabels() {
		if(labelToValueMap == null) {
			return stringLabels;
		}

		return labelToValueMap.keySet();
	}

	/**
	 * Adds a label / value pair to this {@link LabelValueMapping}. This methods ensures that the
	 * label set managed by this {@link LabelValueMapping} unique.
	 */
	private void addLabelValuePair(String label, Object value) {
		// Append IDs for IIdLabeled objects
		if(value instanceof IIdLabeled) {
			label += " [Id=" + ((IIdLabeled)value).getId() + "]";
		}

		// Be extra careful and add unique suffix if required.
		String prefix = new String(label);
		int i = 1;
		while(labelToValueMap.keySet().contains(label)) {
			label = prefix + " (" + i + ")";
			i++;
		}
		label = prefix;

		labelToValueMap.put(label, value);
	}

	/** Translates a value into the corresponding label. */
	public String getLabelForValue(Object value) {
		if(labelToValueMap == null) {
			return value != null ? value.toString() : "";
		}

		for(String key : labelToValueMap.keySet()) {
			Object currValue = labelToValueMap.get(key);

			if(value == currValue) {
				return key;
			}
		}

		// In case the label could not be found, return a string representation of the value. This
		// is useful to implement custom label providers for model elements are edited in combo
		// boxes.
		return value != null ? value.toString() : "";
	}

	/** Translates a label into the corresponding value. */
	public Object getValueForLabel(String label) {
		if(labelToValueMap == null) {
			return label;
		}

		return labelToValueMap.get(label);
	}
}
