/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 ForTISS GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.utils;

import static java.lang.Math.min;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_SHAPE_INSETS;
import static org.fortiss.tooling.base.model.layout.EOrientation.EAST;
import static org.fortiss.tooling.base.model.layout.EOrientation.NORTH;
import static org.fortiss.tooling.base.model.layout.EOrientation.SOUTH;
import static org.fortiss.tooling.base.model.layout.EOrientation.WEST;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.getNodePosition;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.getNodeSize;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.setConnectorOffsetOrientation;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.setNodePosition;
import static org.fortiss.tooling.base.utils.LayoutModelElementFactory.createOffsetOrientation;
import static org.fortiss.tooling.base.utils.SnapToGridUtils.snapToGrid;

import org.fortiss.tooling.base.dnd.ElementDropContext;
import org.fortiss.tooling.base.model.layout.Dimension;
import org.fortiss.tooling.base.model.layout.EOrientation;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.model.layout.OffsetOrientation;
import org.fortiss.tooling.base.model.layout.Point;
import org.fortiss.tooling.kernel.utils.EcoreUtils;

/**
 * Utility class for rectangle-shaped edit part layout.
 * 
 * @author wandinger
 * @author hummel
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 6B711798C123FF6553B50A3053E94CC8
 */
public class RectangleLayoutUtils {
	/**
	 * Returns the {@link OffsetOrientation} of the point on the rectangle
	 * border having a given Point inside the rectangle.
	 * 
	 * @param source
	 *            the relative coordinates in the rectangle
	 * @param rectangleSize
	 *            the size of the rectangle
	 * @param inset
	 *            the inset from each side
	 * @return the offset and orientation
	 */
	public static OffsetOrientation calculateRectangleBorderOrientation(Point source,
			Dimension rectangleSize, int inset) {
		int distToNorth = source.getY();
		int distToSouth = rectangleSize.getHeight() - source.getY() - 2 * inset;
		int distToWest = source.getX();
		int distToEast = rectangleSize.getWidth() - source.getX() - 2 * inset;
		int minDist = min(min(min(distToEast, distToNorth), distToWest), distToSouth);

		int offset;
		EOrientation orientation;
		if(minDist == distToNorth) {
			orientation = NORTH;
			offset = source.getX();
		} else if(minDist == distToWest) {
			orientation = WEST;
			offset = source.getY();
		} else if(minDist == distToSouth) {
			orientation = SOUTH;
			offset = source.getX();
		} else {
			orientation = EAST;
			offset = source.getY();
		}
		return createOffsetOrientation(offset, orientation, null);
	}

	/**
	 * Layouts the given connector within the surrounding rectangle.
	 * 
	 * @param layouted
	 *            the connector element
	 * @param container
	 *            the container of the connector
	 * @param location
	 *            the location in the rectangle to be used as hint
	 */
	public static void layoutConnectorInRectangle(ILayoutedModelElement layouted,
			ILayoutedModelElement container, Point location) {
		// Point position = location.getCopy();
		Point position = EcoreUtils.copy(location);
		org.fortiss.tooling.base.model.layout.Point np = getNodePosition(container);
		position.translate(-np.getX() - DEFAULT_SHAPE_INSETS, -np.getY() - DEFAULT_SHAPE_INSETS);
		// do border clamping
		org.fortiss.tooling.base.model.layout.Dimension nd = getNodeSize(container);
		// Dimension d = new Dimension(nd.getWidth(), nd.getHeight());
		Dimension d =
				LayoutModelElementFactory.createDimension(nd.getWidth(), nd.getHeight(),
						layouted.toString() + "_Dimension");
		OffsetOrientation orient =
				calculateRectangleBorderOrientation(position, d, DEFAULT_SHAPE_INSETS);
		// snap to grid and store the value in the model
		orient.setOffset(snapToGrid(orient.getOffset()));
		setConnectorOffsetOrientation(layouted, orient);
	}

	/**
	 * Layouts the given layouted object using the position in the context
	 * 
	 * @param layouted
	 *            the node to be layouted
	 * @param context
	 *            the location context information to be used
	 * @param positionOffset
	 *            the offset to adj ust the context location, e.g. grid size
	 * @param snapToGrid
	 *            snap to grid should be performed or not
	 */
	public static void layoutNodeInRoot(ILayoutedModelElement layouted, ElementDropContext context,
			int positionOffset, boolean snapToGrid) {
		layoutNodeInRoot(layouted, context.getPosition(), positionOffset, snapToGrid);
	}

	/**
	 * Layouts the given layouted object using the position in the context.
	 * 
	 * @param layouted
	 *            the node to be layouted
	 * @param location
	 *            the location information to be used
	 * @param positionOffset
	 *            the offset to adjust the location, e.g. grid size
	 * @param snapToGrid
	 *            snap to grid should be performed or not
	 */
	public static void layoutNodeInRoot(ILayoutedModelElement layouted, Point location,
			int positionOffset, boolean snapToGrid) {
		Point position = EcoreUtils.copy(location);
		position.translate(-positionOffset, -positionOffset);

		if(snapToGrid) {
			SnapToGridUtils.snapToGrid(position);
		}
		setNodePosition(layouted, position.getX(), position.getY());
	}
}
