/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2015 ForTISS GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.constraint;

import static org.fortiss.tooling.kernel.constraint.ConstraintMessage.createDuplicateIdViolation;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.fortiss.tooling.kernel.extension.IConstraintChecker;
import org.fortiss.tooling.kernel.extension.base.ConstraintCheckerBase;
import org.fortiss.tooling.kernel.model.IIdLabeled;

/**
 * <p>
 * {@link IConstraintChecker} to check whether all {@link IIdLabeled} elements below a given root
 * {@link EObject} actually have unique IDs.
 * </p>
 * <p>
 * While the uniqueness of IDs should be ensured by the assignment strategy of the kernel, and
 * {@code ModelContext} eliminates duplicated IDs when loading and saving models, this constraint
 * checker is an additional safety layer.
 * </p>
 * <p>
 * This is because in persisted model resources, the encoding of {@link EReference} is based on IDs.
 * Hence, model resources with duplicated IDs in referenced model elements are invalid and cannot be
 * loaded.
 * </p>
 * <p>
 * <b>CAVEAT:</b> For performance reasons, be sure to register this constraint checker only for a
 * class the exclusively represents the root of your model.
 * </p>
 * 
 * @author barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 0AB4B0FC7A63A2A975085BFE6C67EBE9
 */
public class IdUniquenessChecker extends ConstraintCheckerBase<EObject> {

	/** {@inheritDoc} */
	@Override
	public List<IdUniquenessConstraintViolation<IIdLabeled>> apply(EObject rootObject) {
		ArrayList<IdUniquenessConstraintViolation<IIdLabeled>> result =
				new ArrayList<IdUniquenessConstraintViolation<IIdLabeled>>();

		Map<Integer, IIdLabeled> idToElementMap = new HashMap<Integer, IIdLabeled>();

		for(IIdLabeled current : getChildrenWithType(rootObject, IIdLabeled.class)) {
			// In case there are duplicates, the first occurrence is considered to be legitimate
			// and is hence kept in the map
			int id = current.getId();
			if(idToElementMap.containsKey(id)) {
				result.add(createDuplicateIdViolation(current, idToElementMap.get(id)));
			} else {
				idToElementMap.put(id, current);
			}
		}

		return result;
	}
}
